use std::{path::PathBuf, sync::Arc};

use eyre::ContextCompat;
use git2::{Cred, PushOptions, RemoteCallbacks, Repository};
use tokio::sync::Mutex;

use crate::storage::DynStorageEngine;

use super::GitProvider;

#[derive(Clone, Debug)]
pub struct LocalGitProviderOptions {
    pub http_auth: Option<String>,
}

pub struct LocalGitProvider {
    storage_engine: DynStorageEngine,
    options: LocalGitProviderOptions,
}

impl LocalGitProvider {
    pub fn new(options: LocalGitProviderOptions, storage_engine: DynStorageEngine) -> Self {
        Self {
            storage_engine,
            options,
        }
    }
}

#[async_trait::async_trait]
impl GitProvider for LocalGitProvider {
    async fn clone_from_url(&self, url: &String) -> eyre::Result<(PathBuf, Repository)> {
        let url = url.clone();
        tracing::debug!(url, "allocating dir");
        let dir = self.storage_engine.allocate_dir().await?;
        let options = self.options.clone();

        let dirpath = dir.clone().path();
        let repo = tokio::task::spawn_blocking(move || {
            let mut callbacks = RemoteCallbacks::new();
            callbacks.credentials(|url, username_from_url, _allowed_types| {
                tracing::debug!(username_from_url, url, "pulling key from ssh-agent");

                if let Some(auth) = &options.http_auth {
                    tracing::trace!(auth, "authenticating");
                    let (user, pass) = auth
                        .split_once(":")
                        .ok_or("http_auth is not formatted correctly")
                        .unwrap();

                    Cred::userpass_plaintext(user, pass)
                } else {
                    let username = username_from_url
                        .context("could not find username_from_url")
                        .unwrap();
                    Cred::ssh_key_from_agent(username)
                }
            });

            let mut fo = git2::FetchOptions::new();
            fo.remote_callbacks(callbacks);

            let checkout_builder = git2::build::CheckoutBuilder::new();

            let mut builder = git2::build::RepoBuilder::new();
            builder.fetch_options(fo).with_checkout(checkout_builder);

            tracing::debug!(
                path = dirpath.as_os_str().to_string_lossy().to_string(),
                "clone git repo"
            );
            builder.clone(url.as_str(), dirpath.as_path())
        })
        .await??;

        tracing::debug!("done pulling repo");

        Ok((dir.path(), repo))
    }

    async fn create_branch(
        &self,
        repo: Arc<Mutex<Repository>>,
        branch_name: &String,
    ) -> eyre::Result<()> {
        let repo = repo.lock().await;

        let head_commit_oid = repo
            .head()?
            .target()
            .ok_or(eyre::anyhow!("could not get access to target commit"))?;
        let head_commit = repo.find_commit(head_commit_oid)?;
        let newbranch = repo.branch(
            &branch_name.to_lowercase().replace(" ", "-"),
            &head_commit,
            true,
        )?;

        repo.set_head(
            newbranch
                .into_reference()
                .name()
                .ok_or(eyre::anyhow!("could not get name of reference"))?,
        )?;

        Ok(())
    }

    async fn push_branch(
        &self,
        repo: Arc<Mutex<Repository>>,
        branch_name: &String,
    ) -> eyre::Result<()> {
        let repo = repo.lock().await;
        let options = self.options.clone();

        tracing::trace!("pulling signature from local git");
        let signature = repo.signature()?;

        for rev in repo.revwalk()? {
            let rev = rev?;
            let commit = repo.find_commit(rev)?;
            let summary = commit
                .summary()
                .ok_or(eyre::anyhow!("could not find commit"))?;
            tracing::info!(summary, "summary")
        }

        tracing::trace!("fetching index and adding changed files to working tree");
        let mut index = repo.index()?;
        index.add_all(&["."], git2::IndexAddOption::DEFAULT, None)?;
        index.write()?;
        let tree = index.write_tree()?;
        let tree = repo.find_tree(tree)?;

        let parents = repo.head().map(|h| {
            h.target()
                .ok_or(eyre::anyhow!("could not fetch target"))
                .map(|t| repo.find_commit(t))
        })???;

        tracing::trace!("writing commit object");
        repo.commit(
            Some("HEAD"),
            &signature,
            &signature,
            branch_name.to_lowercase().replace(" ", "-").as_str(),
            &tree,
            &[&parents],
        )?;

        let mut remote = repo.find_remote("origin")?;
        let head = repo.head()?;
        let refspec = &[head
            .name()
            .ok_or(eyre::anyhow!("could not find head.name"))?];

        let mut remote_callbacks = RemoteCallbacks::new();
        remote_callbacks.credentials(|url, username_from_url, _allowed_types| {
            tracing::debug!(username_from_url, url, "pulling key from ssh-agent");

            if let Some(auth) = &options.http_auth {
                tracing::trace!(auth, "authenticating");
                let (user, pass) = auth
                    .split_once(":")
                    .ok_or("http_auth is not formatted correctly")
                    .unwrap();

                Cred::userpass_plaintext(user, pass)
            } else {
                let username = username_from_url.unwrap();
                Cred::ssh_key_from_agent(username)
            }
        });

        let mut push_options = PushOptions::new();
        push_options.remote_callbacks(remote_callbacks);

        tracing::trace!("pushing to remote");
        remote.push(refspec, Some(&mut push_options))?;

        Ok(())
    }
}
