use std::{path::PathBuf, sync::Arc};

use async_trait::async_trait;

use super::models::Schema;

#[async_trait]
pub trait SchemaParser {
    async fn parse_file(&self, file: PathBuf) -> eyre::Result<Schema>;
}

pub type DynSchemaParser = Arc<dyn SchemaParser + Send + Sync>;

#[derive(Debug)]
pub struct DefaultSchemaParser {}

#[async_trait]
impl SchemaParser for DefaultSchemaParser {
    async fn parse_file(&self, file: PathBuf) -> eyre::Result<Schema> {
        let file = tokio::fs::read(file).await?;

        self.parse(file)
    }
}

impl DefaultSchemaParser {
    pub fn new() -> Self {
        Self {}
    }

    pub fn parse(&self, contents: Vec<u8>) -> eyre::Result<Schema> {
        let schema = serde_yaml::from_slice(contents.as_slice())?;

        Ok(schema)
    }
}

mod test {
    use super::DefaultSchemaParser;
    use crate::schema::models::{Action, Schema, SelectAction};

    #[test]
    fn can_parse_action() {
        let content = r#"apiVersion: action
name: write-a-readme
select:
  repositories:
    - git@git.front.kjuulh.io:kjuulh/octopush-test.git
actions:
  - type: go
    entry: "main.go"
"#;

        let res = DefaultSchemaParser::new().parse(content.trim().into());

        assert_eq!(
            res.unwrap(),
            Schema::Action {
                name: "write-a-readme".into(),
                select: SelectAction {
                    repositories: vec!["git@git.front.kjuulh.io:kjuulh/octopush-test.git".into()]
                },
                action: vec![Action::Go {
                    entry: "main.go".into()
                }]
            }
        )
    }
}
